defmodule ClickHouse.ResultStream do
  @moduledoc """
  A way to manage the results of a ClickHouse query stream.
  """

  @typedoc """
  The state of a ClickHouse query stream.
  """
  @type t :: %__MODULE__{
          id: reference(),
          pid: pid(),
          meta: map(),
          done: boolean(),
          chunks: integer()
        }

  @derive {Inspect, only: [:done, :chunks]}
  defstruct [:id, :pid, :meta, done: false, chunks: 0]
end

defmodule ClickHouse.ResultStreamChunk do
  @moduledoc """
  A single chunk of data returned by a ClickHouse query stream.
  """

  @typedoc """
  A single chunk of data from a ClickHouse query stream.
  """
  @type t :: %__MODULE__{
          id: reference(),
          data: iodata()
        }

  @derive {Inspect, only: []}
  defstruct [:id, :data]
end

defmodule ClickHouse.Stream do
  @moduledoc false

  alias ClickHouse.{ResultStream, ResultStreamChunk}

  ################################
  # Public API
  ################################

  @doc false
  @spec new(
          ClickHouse.client(),
          ClickHouse.Query.t(),
          ClickHouse.stream_handler(),
          opts :: keyword()
        ) ::
          Enumerable.t()
  def new(client, query, handler, opts) do
    Stream.resource(
      fn -> stream_start(client, query, opts) end,
      fn acc -> stream_next(client, acc, handler) end,
      fn acc -> acc end
    )
  end

  ################################
  # Private API
  ################################

  defp stream_start(client, query, opts) do
    client.interface.stream(client, query, opts)
  end

  defp stream_next(client, {:ok, result_stream}, handler) do
    case client.interface.stream_next(client, result_stream) do
      {:ok, %ResultStream{done: false} = result_stream} ->
        receive do
          %ResultStreamChunk{} = chunk ->
            handler.(chunk)
            {[chunk], {:ok, result_stream}}
        end

      {:ok, _} = success ->
        {:halt, success}

      {:error, _} = error ->
        {:halt, error}
    end
  end
end
